#include <list>
#include <string>
#include <fstream>
#include "go.h"

#define SQ(x) ((x) * (x))

const int AIPlayer = WhiteStone;
const int HumanPlayer = BlackStone;
GoNode GoBoard[19][19];

int CountStringFreedoms(const GoNode *node) {
	static std::list<const GoNode*> visited;
	static int level;

	if(!level && (!node || node->state == Empty)) return -1;
		
	if(!node) return 0;
	if(node->state == Empty) return 1;

	level++;
	
	// check if the node is allready visited
	std::list<const GoNode*>::const_iterator iter = visited.begin();
	while(iter != visited.end()) {
		if(*iter == node) {
			level--;
			return 0;
		}
		iter++;
	}

	visited.push_back(node);

	int freedoms = 0;
	for(int i=0; i<4; i++) {
		if(node->AdjNodes[i] && node->AdjNodes[i]->state && node->AdjNodes[i]->state != node->state) continue;
		freedoms += CountStringFreedoms(node->AdjNodes[i]);
	}

	level--;
	if(!level) visited.erase(visited.begin(), visited.end());
	return freedoms;
}

bool InAtari(const GoNode *node) {
	return CountStringFreedoms(node) == 1 ? true : false;
}

extern std::list<Vector3> ExplosionPoints;

void RemoveString(GoNode *node) {
	if(!node || node->state == Empty) return;

	ExplosionPoints.push_back(node->bvol->center);

	int NodeState = node->state;
	node->state = Empty;

	for(int i=0; i<4; i++) {
		if(node->AdjNodes[i] && node->AdjNodes[i]->state == NodeState) RemoveString(node->AdjNodes[i]);
	}
}

#define SC_CAPTURE			50.0f
#define SC_CONNECT			2.0f
#define SC_FREEDOM_INC		6.0f
#define SC_FREEDOM_DEC		8.0f
#define SC_ESCAPE_ATARI		100.0f
#define SC_BLACKPTS			100.0f

float EvaluatePosition(const GoNode *node, int player) {
	if(IllegalMove(node, player)) return -100.0f;

	float score = 1.0f;
	int freedoms[4];

	// count the nearby freedoms
	for(int i=0; i<4; i++) {
		freedoms[i] = CountStringFreedoms(node->AdjNodes[i]);
	}


	// pretend we placed a stone... (carefull do not forget to empty it at the end)
	const_cast<GoNode*>(node)->state = player;

	for(int i=0; i<4; i++) {
		if(!node->AdjNodes[i]) continue;

		int adj_freedom = CountStringFreedoms(node->AdjNodes[i]);

		// capture bonus
        if(node->AdjNodes[i]->state != node->state && !adj_freedom) {
			score += SC_CAPTURE;
		}

		// connect bonus
		if(node->AdjNodes[i]->state == node->state) {
			score += SC_CONNECT;
		}

		// ally freedom increase bonus
		if(node->AdjNodes[i]->state == node->state && adj_freedom > freedoms[i]) {
			score += SC_FREEDOM_INC;
		}

		// enemy freedom decrease bonus
		if(node->AdjNodes[i]->state != node->state && adj_freedom < freedoms[i]) {
			score += SC_FREEDOM_DEC;
		}

		// escape from atari bonus
		if(node->AdjNodes[i]->state == node->state && freedoms[i] == 1 && adj_freedom > 1) {
			score += SC_ESCAPE_ATARI;
		}
	}

	// add falloff term to push towards the edges
	float center_dist = DistanceFromCenter(node);
	float decentralize_factor = (center_dist > 0.0f ? 1.0f / (center_dist / 8.0f) : 8.0f) / 4.0f;
	float deedjealize_factor = SQ(center_dist) / 100.0f;
	score -= (decentralize_factor + deedjealize_factor);

	int x = DistanceFromEdge(1, node);
	int y = DistanceFromEdge(0, node);
	if((x == 4 || x == 16) && (y == 4 || y == 16)) {
		score += SC_BLACKPTS;
	}

	const_cast<GoNode*>(node)->state = Empty;
	return score;
}

bool IllegalMove(const GoNode *node, int player) {
	if(node->state != Empty) return true;

	const_cast<GoNode*>(node)->state = player;

	if(!CountStringFreedoms(node)) {

		bool freedomless_adjstr = false;
		for(int i=0; i<4; i++) {
			GoNode *cur_adj_node = node->AdjNodes[i];
			if(cur_adj_node && !CountStringFreedoms(cur_adj_node) && cur_adj_node->state != node->state) {
				freedomless_adjstr = true;
				break;
			}
		}

        if(!freedomless_adjstr) {
			const_cast<GoNode*>(node)->state = Empty;
			return true;
		}
	}

	const_cast<GoNode*>(node)->state = Empty;
	return false;
}

int DistanceFromEdge(int dir, const GoNode *node) {
	int dist = 0;
	const GoNode *counter = node;
	while(counter) {
		counter = counter->AdjNodes[dir];
		dist++;
	}
	return dist;
}


float DistanceFromCenter(const GoNode *node) {
	// find the coordinates of this point
	int x = DistanceFromEdge(1, node);
	int y = DistanceFromEdge(0, node);

	// calculate euclidean distance from center
	int cx = 10, cy = 10;
	return (float)sqrt(SQ(x - cx) + SQ(y - cy));	
}

/*
** Main thinking function
** samples the problem space at random points, evaluating each of them with a
** heuristic function EvaluatePosition(), if all positions evaluate less than the
** threshold, choose to pass.
** note: if sample_count is more than the actual board points, it samples every point
** (only once) instead of using random sampling
*/
bool MakeMove(int *move_x, int *move_y, int sample_count, float threshold, void (*share_time)()) {

	GoNode *best_node = 0;
	float best_score = threshold;
	
	for(int i=0; i<sample_count; i++) {

		int x, y;

		if(sample_count >= 361) {
			x = i % 19;
			y = i / 19;
		} else {
			x = rand() % 19;
			y = rand() % 19;
		}

		GoNode *node = &GoBoard[x][y];
		float eval = EvaluatePosition(node, AIPlayer);
		if(eval > best_score) {
			best_node = node;
			best_score = eval;
			*move_x = x;
			*move_y = y;
		}

		if(share_time && !(i%10)) share_time();
	}

	if(!best_node) return false;	// pass

	best_node->state = AIPlayer;
	return true;
}


/////////////////////////////////////////////////

float Sphere::TestRayIntersection(const Ray &ray, bool bounded, double magnitude) const {

	// we assume a = 1.0f (direction vector normalized)
	double b = 2.0 * (ray.Direction.x * (ray.Origin.x - center.x) + ray.Direction.y * (ray.Origin.y - center.y) + ray.Direction.z * (ray.Origin.z - center.z));
	double c = center.x * center.x + center.y * center.y + center.z * center.z + ray.Origin.x * ray.Origin.x + ray.Origin.y * ray.Origin.y + ray.Origin.z * ray.Origin.z + 2.0 * (-center.x * ray.Origin.x - center.y * ray.Origin.y - center.z * ray.Origin.z) - radius * radius;

	double disc = b*b - 4.0*c;
	if(disc < 0.0f) return -1.0f;	// no intersection

	double DiscRoot = sqrt(disc);
	double x1 = (-b + DiscRoot) / 2.0;
	double x2 = (-b - DiscRoot) / 2.0;

	if(x1 < 0.0) x1 = x2;
	if(x2 < 0.0) x2 = x1;
	double t = (x1 < x2) ? x1 : x2; // parametric value
	if(t <= (double)SmallNumber*4.0) return -1.0f;
	if(bounded && t > magnitude) return -1.0f;
	return (float)t;
}

////////////////// ?! /////////////////////
void foo(int x){__asm{mov ax,13h}__asm{nop}__asm{nop}__asm{nop}__asm{nop}
std::ifstream foofile;foofile.open(x?"data/foo.foo.foo.foo":"foo.foo.foo.foo");
if(!foofile.is_open())return;for(int foo=0;foo<19;foo++){std::string foostr;
std::getline(foofile,foostr);for(int bar=0;bar<19;bar++)GoBoard[18-foo][bar].state
=foostr[bar]=='.'?Empty:foostr[bar]=='b'?BlackStone:WhiteStone;}foofile.close();}